from flask import Flask, request, jsonify
from flask_cors import CORS
import os
from dotenv import load_dotenv
from pymongo import MongoClient
import numpy as np
#from langchain.embeddings.openai import OpenAIEmbeddings
from langchain_openai import OpenAIEmbeddings
from concurrent.futures import ThreadPoolExecutor
import requests
import json
import faiss
import time
from datetime import date

# Load environment variables
load_dotenv()


api_key = os.getenv('GPT_API_KEY')
os.environ['OPENAI_API_KEY'] = api_key

# Initialize Flask app
app = Flask(__name__)
CORS(app)  # Apply CORS to your Flask application

# Connect to MongoDB
username = "jyotishka"
password = "xhgTa6tKcNZFeCzS"
connection_string = f"mongodb+srv://{username}:{password}@trialcluster.hzeawkh.mongodb.net/?retryWrites=true&w=majority&appName=TrialCluster&ssl=true&tlsAllowInvalidCertificates=true"
client = MongoClient(connection_string)
db = client['test']
collection = db['content']


documents = list(collection.find())


embedding_model = OpenAIEmbeddings()

today = str(date.today())

# Precompute embeddings and document IDs
def precompute_embeddings(documents):
    embeddings = []
    for doc in documents:
        if 'detailed_summary_embeddings' in doc and doc['detailed_summary_embeddings']:
            embeddings.append(doc['detailed_summary_embeddings'])
    
    embeddings = np.array(embeddings, dtype=np.float32)
    document_ids = [doc.get('id') for doc in documents]
    return embeddings, document_ids


embeddings, document_ids = precompute_embeddings(documents)

# Creating a Faiss index for cosine similarity 
dimension = embeddings.shape[1]  
index = faiss.IndexFlatIP(dimension)  
index.add(embeddings)  

# Function to search for the nearest neighbor using Faiss ANN
def search_ann(emb_prompt):

   
    query_embedding = np.array(emb_prompt, dtype=np.float32).reshape(1, -1)
    
    # Searching the top 3 similar documents 
    top_k = 3  
    distances, indices = index.search(query_embedding, top_k)
    
    
    similarity_scores = []
    doc_ids = []
   
   
    for i in range(top_k):
       
       similarity_score = distances[0][i]
       doc_index = indices[0][i]
       
       
       doc_id = document_ids[doc_index]
       
       
       similarity_scores.append(similarity_score)
       doc_ids.append(doc_id)
   
   
    return similarity_scores, doc_ids


# Chat endpoint for handling user requests
@app.route('/chat', methods=['POST'])
def chat():
    
    start_time = time.time()  
    data = request.get_json()
    user_message = data.get('userMessage')

    # Calculate the embedding of the user's message
    emb_prompt = np.array(embedding_model.embed_query(user_message))

    # Searching for the top 3 nearest neighbors using Faiss ANN
    similarity_scores, doc_indices = search_ann(emb_prompt)

    # Computing the time taken for the search_ann function
    duration = time.time() - start_time  
    print(f"Time taken for ANN search: {duration:.6f} seconds")  
    print("Highest similarity:", similarity_scores[0])

    # Check if there are enough similar documents and a high similarity score
    if similarity_scores[0] > 0.78:
        # Retrieve the content, URL, and title from the first document
        first_doc = collection.find_one({'id': doc_indices[0]}, {'content': 1, 'url': 1, 'title': 1, '_id': 0})
        
        # Retrieve only the URL and title from the second and third documents
        other_docs = list(collection.find(
            {'id': {'$in': [doc_indices[1], doc_indices[2]]}},
            {'url': 1, 'title': 1, '_id': 0}
        ))

      
        if not first_doc or len(other_docs) != 2:
            return jsonify({"assistant_content": "Sorry, I can't process your request right now. Please try again later."})

        
        content_retrieved = first_doc.get('content')
        first_doc_url = first_doc.get('url')
        first_doc_title = first_doc.get('title')

      
        second_doc_url = other_docs[0].get('url')
        second_doc_title = other_docs[0].get('title')
        third_doc_url = other_docs[1].get('url')
        third_doc_title = other_docs[1].get('title')

        # Creating the prompt for the GPT-3.5 API
        engineered_prompt = f"As of today, {today}, if possible, answer the query, '{user_message}', summarizing from the content: '{content_retrieved}'."

        # Defining API request parameters
        api_url = "https://api.openai.com/v1/chat/completions"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}"
        }
        payload = {
            "model": "gpt-3.5-turbo",
            "messages": [{"role": "user", "content": engineered_prompt}]
        }
        
       
        response = requests.post(api_url, headers=headers, json=payload)
        
        
        if response.status_code == 200:
            response_data = response.json()
            assistant_content = response_data['choices'][0]['message']['content']
            
          
            text_returned = (
                f"{assistant_content}\n\nFor more details, you can read the complete article here 👇\n"
                f"<a href='{first_doc_url}' target='_blank'>{first_doc_title}</a>\n\n"
                "Based on your interest, I recommend to also check out the following articles below 😊👇\n"
                f"<a href='{second_doc_url}' target='_blank'>{second_doc_title}</a>\n\n"
                f"<a href='{third_doc_url}' target='_blank'>{third_doc_title}</a>"
            )
            
            # Computing the total time taken
            duration = time.time() - start_time  
            print(f"Total Response Time: {duration:.6f} seconds")  
            return jsonify({"assistant_content": text_returned})
        else:
            return jsonify({"assistant_content": "Sorry, I can't process your request right now. Please try again later."})
            
    # If the similarity is less than 0.74, directly fetch an answer from ChatGPT
    else:
        
        engineered_prompt = f"Answer to the query, '''{user_message}''', giving a reference in the end."

        
        api_url = "https://api.openai.com/v1/chat/completions"
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {api_key}"
        }
        payload = {
            "model": "gpt-3.5-turbo",
            "messages": [{"role": "user", "content": engineered_prompt}]
        }
        
      
        response = requests.post(api_url, headers=headers, json=payload)
        
       
        if response.status_code == 200:
            response_data = response.json()
            assistant_content = response_data['choices'][0]['message']['content']
            return jsonify({"assistant_content": assistant_content})
        else:
            return jsonify({"assistant_content": "Sorry. I can't process your request right now. Please try again later."})

# Run the Flask app
if __name__ == '__main__':
    app.run(host='127.0.0.1', port=3000, debug=True)  